//
//  ViewController.m
//  usb-rly82
//
//  Created by James Henderson on 28/09/2016.
//  Copyright © 2016 Devantech Ltd. All rights reserved.
//

#import "ViewController.h"

@implementation ViewController

- (void)viewDidLoad {
    [super viewDidLoad];
    // this notification listens for when you click on the serial drop down and calls refreshSerialList
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(refreshSerialList:) name:@"NSPopUpButtonWillPopUpNotification" object:nil];
}

- (void)setRepresentedObject:(id)representedObject {
    [super setRepresentedObject:representedObject];
    
    // Update the view, if already loaded.
}

- (void) viewDidDisappear {
    [super viewDidDisappear];
    [self closePort];
}

/*
 *
 *  Pop up button functions.
 *
 */

- (IBAction)voltageRefSelected:(id)sender
{
    if ([sender isKindOfClass:[NSPopUpButton class]]) {
        NSPopUpButton *pop = (NSPopUpButton *)sender;
        unsigned char tag = (unsigned char)pop.tag;                  // The tag from the control will give us wich adc chanel we wish to set the reference for (1-8)
        unsigned char sel = (unsigned char)pop.indexOfSelectedItem;  // The selected index will give us the reference voltage we wish to set (0-2)
        
        // Put the command together to set the adv reference
        serBuf[0] = 0x81;
        serBuf[1] = tag;
        serBuf[2] = sel;
        [self transmit:3];
    }
}

- (void)setRefVoltageIndexes
{
    [adc1ref selectItemAtIndex:adcref1];
    [adc2ref selectItemAtIndex:adcref2];
    [adc3ref selectItemAtIndex:adcref3];
    [adc4ref selectItemAtIndex:adcref4];
    [adc5ref selectItemAtIndex:adcref5];
    [adc6ref selectItemAtIndex:adcref6];
    [adc7ref selectItemAtIndex:adcref7];
    [adc8ref selectItemAtIndex:adcref8];
}

// Called when serial port selected from serial port list
- (IBAction)serialPortSelected:(id)sender{
    [self closePort];
    if ([self openPort:[serialList titleOfSelectedItem]]){
        
        // Get the firmware version and module id
        serBuf[0] = 0x5A;
        [self transmit:1];
        [self receive:2];
        
        // Display the firmware version and module id
        NSString *verString = [NSString stringWithFormat:@"Firmware v : %d", serBuf[1]];
        NSString *idString = [NSString stringWithFormat:@"Module id : %d", serBuf[0]];
        [firmwareLabel setStringValue:verString];
        [idLabel setStringValue:idString];
        
        // Get the relay states
        serBuf[0] = 0x5B;
        [self transmit:1];
        [self receive:1];
        relaystates = serBuf[0];
        
        [self setButtonColours];
        
        // get the adc references
        serBuf[0] =0x82;
        [self transmit:1];
        [self receive:8];
        
        adcref1 = serBuf[0];
        adcref2 = serBuf[1];
        adcref3 = serBuf[2];
        adcref4 = serBuf[3];
        adcref5 = serBuf[4];
        adcref6 = serBuf[5];
        adcref7 = serBuf[6];
        adcref8 = serBuf[7];
        
        [self setRefVoltageIndexes];
        
        // Get the adc values
        serBuf[0] = 0x80;
        [self transmit:1];
        [self receive:16];
        
        adc1 = serBuf[0];   // Put tthe adc values together
        adc1 <<= 8;
        adc1 += serBuf[1];
        adc2 = serBuf[2];
        adc2 <<= 8;
        adc2 += serBuf[3];
        adc3 = serBuf[4];
        adc3 <<= 8;
        adc3 += serBuf[5];
        adc4 = serBuf[6];
        adc4 <<= 8;
        adc4 += serBuf[7];
        adc5 = serBuf[8];
        adc5 <<= 8;
        adc5 += serBuf[9];
        adc6 = serBuf[10];
        adc6 <<= 8;
        adc6 += serBuf[11];
        adc7 = serBuf[12];
        adc7 <<= 8;
        adc7 += serBuf[13];
        adc8 = serBuf[14];
        adc8 <<= 8;
        adc8 += serBuf[15];
        
        NSString *adc1string = [NSString stringWithFormat:@"ADC1 : %d",adc1];
        NSString *adc2string = [NSString stringWithFormat:@"ADC2 : %d",adc2];
        NSString *adc3string = [NSString stringWithFormat:@"ADC3 : %d",adc3];
        NSString *adc4string = [NSString stringWithFormat:@"ADC4 : %d",adc4];
        NSString *adc5string = [NSString stringWithFormat:@"ADC5 : %d",adc5];
        NSString *adc6string = [NSString stringWithFormat:@"ADC6 : %d",adc6];
        NSString *adc7string = [NSString stringWithFormat:@"ADC7 : %d",adc7];
        NSString *adc8string = [NSString stringWithFormat:@"ADC8 : %d",adc8];
        [adc1Label setStringValue:adc1string];
        [adc2Label setStringValue:adc2string];
        [adc3Label setStringValue:adc3string];
        [adc4Label setStringValue:adc4string];
        [adc5Label setStringValue:adc5string];
        [adc6Label setStringValue:adc6string];
        [adc7Label setStringValue:adc7string];
        [adc8Label setStringValue:adc8string];
        
        relay1Button.enabled = true;
        relay2Button.enabled = true;
        adc1ref.enabled = true;
        adc2ref.enabled = true;
        adc3ref.enabled = true;
        adc4ref.enabled = true;
        adc5ref.enabled = true;
        adc6ref.enabled = true;
        adc7ref.enabled = true;
        adc8ref.enabled = true;
        
        [NSTimer scheduledTimerWithTimeInterval:0.1 target:self selector:@selector(timerTick:) userInfo:NULL repeats:NO];
    }
}

- (void) refreshSerialList:(NSNotification*) notification{
    io_object_t sPort;
    io_iterator_t serialPortIterator;
    
    NSString *selectedItem = [serialList titleOfSelectedItem];       // Store selected item so it can be reselected
    
    [serialList removeAllItems];                                     // Clear entries in serial port list
    
    IOServiceGetMatchingServices(kIOMasterPortDefault, IOServiceMatching(kIOSerialBSDServiceValue), &serialPortIterator);   // ask for all serial ports
    
    while ((sPort = IOIteratorNext(serialPortIterator))) {          // loop through ports and add them to an array
        [serialList addItemWithTitle:(__bridge NSString*)IORegistryEntryCreateCFProperty(sPort, CFSTR(kIOCalloutDeviceKey),  kCFAllocatorDefault, 0)];
        IOObjectRelease(sPort);
    }
    
    [serialList insertItemWithTitle:@"Select Serial Port" atIndex:0];
    [serialList selectItemWithTitle:selectedItem];
    if (serialList.indexOfSelectedItem < 0) [serialList selectItemAtIndex:0];
    IOObjectRelease(serialPortIterator);
}

/*
 *
 *  Button functions
 *
 */

- (void) buttonPressed:(id)sender
{
    if ([sender isKindOfClass:[NSButton class]]) {              // Make sure the control sending the message is a button
        NSButton *button = (NSButton*)sender;
        NSString *st = button.title;                            // Get the title of the button
        char relay = [st characterAtIndex:(st.length - 1)];     // Get the last char of the title which is the relay number
        switch (relay) {                                        // Find which relay we wish to change and put the command in serBuf[0] to either turn it on or off
            case '1':                                   // Relay 1
                if (!(relaystates & 0x01)) {            // Check state of relay
                    serBuf[0] = 0x65;                   // Relay 1 on
                } else {
                    serBuf[0] = 0x6F;                   // Relay 1 off
                }
                break;
            case '2':
                if (!(relaystates & (0x01 << 1))) {
                    serBuf[0] = 0x66;
                } else {
                    serBuf[0] = 0x70;
                }
                break;
        }
    }
    [self transmit:1];      // Send the command
}

- (void) setButtonColours
{
    if (relaystates & 0x01) {
        [relay1Button setState:1];
    } else {
        [relay1Button setState:0];
    }
    
    if (relaystates & (0x01 << 1)) {
        [relay2Button setState:1];
    } else {
        [relay2Button setState:0];
    }
}

/*
 *
 *  Timer functions.
 *
 */

- (void) timerTick:(NSTimer *)t
{
    if (isConnected == TRUE) {
        
        // Get the relay states
        serBuf[0] = 0x5B;
        [self transmit:1];
        [self receive:1];
        relaystates = serBuf[0];
        
        [self setButtonColours];
        
        // Get the adc values
        serBuf[0] = 0x80;
        [self transmit:1];
        [self receive:16];
        
        adc1 = serBuf[0];   // Put tthe adc values together
        adc1 <<= 8;
        adc1 += serBuf[1];
        adc2 = serBuf[2];
        adc2 <<= 8;
        adc2 += serBuf[3];
        adc3 = serBuf[4];
        adc3 <<= 8;
        adc3 += serBuf[5];
        adc4 = serBuf[6];
        adc4 <<= 8;
        adc4 += serBuf[7];
        adc5 = serBuf[8];
        adc5 <<= 8;
        adc5 += serBuf[9];
        adc6 = serBuf[10];
        adc6 <<= 8;
        adc6 += serBuf[11];
        adc7 = serBuf[12];
        adc7 <<= 8;
        adc7 += serBuf[13];
        adc8 = serBuf[14];
        adc8 <<= 8;
        adc8 += serBuf[15];
        
        // Display the adc values
        NSString *adc1string = [NSString stringWithFormat:@"ADC1 : %d",adc1];
        NSString *adc2string = [NSString stringWithFormat:@"ADC2 : %d",adc2];
        NSString *adc3string = [NSString stringWithFormat:@"ADC3 : %d",adc3];
        NSString *adc4string = [NSString stringWithFormat:@"ADC4 : %d",adc4];
        NSString *adc5string = [NSString stringWithFormat:@"ADC5 : %d",adc5];
        NSString *adc6string = [NSString stringWithFormat:@"ADC6 : %d",adc6];
        NSString *adc7string = [NSString stringWithFormat:@"ADC7 : %d",adc7];
        NSString *adc8string = [NSString stringWithFormat:@"ADC8 : %d",adc8];
        [adc1Label setStringValue:adc1string];
        [adc2Label setStringValue:adc2string];
        [adc3Label setStringValue:adc3string];
        [adc4Label setStringValue:adc4string];
        [adc5Label setStringValue:adc5string];
        [adc6Label setStringValue:adc6string];
        [adc7Label setStringValue:adc7string];
        [adc8Label setStringValue:adc8string];
        
        [NSTimer scheduledTimerWithTimeInterval:0.1 target:self selector:@selector(timerTick:) userInfo:NULL repeats:NO];
    }
}

/*
 *
 *  Serial port functions.
 *
 */

- (BOOL) openPort:(NSString*)portName
{
    
    struct termios options;
    
    const char *path = [portName UTF8String];
    
    fd = open(path , O_RDWR | O_NOCTTY | O_NONBLOCK);   // open the port
    
    if (fd == -1) {
        [self showAlertWithString:[NSString stringWithFormat:@"Error opening port : %s(%d).\n", strerror(errno), errno]];
        isConnected = FALSE;
        return FALSE;
    }
    
    
    if (ioctl(fd, TIOCEXCL) == -1)                      // Stop repeated opens on this port
    {
        [self showAlertWithString:@"Error setting TIOCEXCL\n"];
        close(fd);
        isConnected = FALSE;
        return FALSE;
    }
    
    if (fcntl(fd, F_SETFL, 0) == -1)                    // Make the port blocking
    {
        [self showAlertWithString:@"Error clearing O_NONBLOCK \n"];
        close(fd);
        isConnected = FALSE;
        return FALSE;
    }
    
    if (tcgetattr(fd, &originalOptions) == -1)          // Get the ports original options
    {
        [self showAlertWithString:@"Error getting tty attributes \n"];
        close(fd);
        isConnected = FALSE;
        return FALSE;
    }
    
    
    options = originalOptions;
    
    cfmakeraw(&options);
    
    cfsetspeed(&options, B19200);                       // Set 19200 baud
    cfsetospeed(&options, B19200);
    
    options.c_cflag |= (CLOCAL | CREAD);				// Enable the receiver and set local mode
    options.c_cflag &= ~PARENB;                         // No parity bit
    
    options.c_cflag |= CSTOPB;                          // Set 2 stop bits
    
    options.c_cflag &= ~CSIZE;                          // Set the character size
    options.c_cflag |= CS8;
    
    if (tcsetattr(fd, TCSANOW, &options) == -1)
    {
        [self showAlertWithString:@"Error setting tty attributes \n"];
        close(fd);
        isConnected = FALSE;
        return FALSE;
    }
    
    isConnected = TRUE;
    
    return isConnected;
    
}

- (void) closePort                                      // Reset port options and close port
{
    if (isConnected) {
        if (tcsetattr(fd, TCSANOW, &originalOptions) == -1)
        {
            [self showAlertWithString:@"SerialController --> closePort : Error resetting tty attributes"];
        }
        isConnected = FALSE;
        close(fd);
    }
}

- (void) transmit:(int)bytes
{
    char result = write(fd, serBuf, bytes);
    tcdrain(fd); // wait for all data to be written
    if (result == -1) {
        [self showAlertWithString:@"Error writing to device"];
    }
}

- (void) receive:(int)bytes
{
    size_t result = read(fd, serBuf, bytes);
    if (result != bytes) {
        [self showAlertWithString:@"Error reading data"];
    }
}

/*
 *
 *  Error alert box
 *
 */

- (void) showAlertWithString:(NSString*)title
{
    NSAlert *alert = [[NSAlert alloc] init];
    [alert addButtonWithTitle:@"OK"];
    [alert setMessageText:title];
    [alert setAlertStyle:NSCriticalAlertStyle];
    [alert runModal];
    [self closePort];
    
}

@end
